#include <iostream>
#include <fstream>
#include <string>
#include <assert.h>
#include <vector>
#include <cstdlib>
#include <complex>
#include <fftw3.h>
#include "reading.hpp"
#include "trace_data.hpp"
#include "trace.hpp"
#include "beamforming.hpp"
#include "filter_workflows.hpp"
using namespace std;

int main(int argc, char* argv[]) {
  try{
    // input file names
    char *fileListName = argv[1]; // input file with list of file names to use
    ifstream fileList(fileListName);
    string filenameStr;

    // y velocities of interest
    vector<float> yvelocities;
    float minvel = atof(argv[2]);
    float maxvel = atof(argv[3]);
    float dvel = atof(argv[4]); 
    float vel = minvel;
    while(vel <= maxvel){
      yvelocities.push_back(vel);
      vel += dvel;
      if(abs(vel) < 0.000001){ // make sure 0 velocity isn't used
	vel += dvel;
      }
    }
    int nvelocities = yvelocities.size();
    // x velocity of interest
    float xvelocity = 0.0;

    // channels of interest and center location in array
    int startChannel = atoi(argv[5]);
    int nChannels = atoi(argv[6]); 
    float xcenter = 0.0;
    float ycenter = float(startChannel + nChannels/2.0);
    
    // open the output text file where beamforming results will be printed
    char *outfileName = argv[7];
    ofstream outfile(outfileName);

    // min and max frequencies to include in signal
    float minFrq = atof(argv[8]);
    float maxFrq = atof(argv[9]);
    cout << minFrq << "\t" << maxFrq << "\n";
    
    // first print all the velocities in the output file
    for(int iv=0; iv<nvelocities; ++iv){
    	outfile << yvelocities[iv] << "\t";
    }
    outfile << "\n";

    if(fileList.is_open() && outfile.is_open()){ 
    
        while(getline(fileList,filenameStr)){
	
		// read relevant traces from this file
		const char *filename = filenameStr.c_str();
		cout << filename << "\n";
		header_vals ThisFileInfo = SEGYInfo(filename); // get nTraces, sampleFrq, nSamples and store in ThisFileInfo
		cout << ThisFileInfo.nTraces << " " << ThisFileInfo.sampleFrq << " " << ThisFileInfo.nSamples << "\n";
		// allocate space for beamforming time series result
		vector<double> beamformResult;
		beamformResult.reserve(ThisFileInfo.nSamples);
		// allocate space for beamforming energy sum result
		vector<double> beamformEnergy;
		beamformEnergy.reserve(nvelocities);
		vector<trace*> traces; // creates a vector of traces
		traces.reserve(nChannels); // preallocate space
		double * data = new double[nChannels*ThisFileInfo.nSamples];
		for(int channel = startChannel; channel < startChannel+nChannels; ++channel){
		    // read a single trace
		    readTrace(filename, &data[(channel-startChannel)*ThisFileInfo.nSamples], channel, ThisFileInfo.nSamples, 0, ThisFileInfo.nSamples);
		    float y = float(channel);
		    traces.push_back(new trace(ThisFileInfo.sampleFrq, ThisFileInfo.nSamples, 0.0, y, &data[(channel-startChannel)*ThisFileInfo.nSamples]));

		    hardBandpass(traces[channel-startChannel], minFrq, maxFrq);
		}

		
		// for each velocity of interest do beamforming and record energy of beamformed time series in vector
		for(int iv=0; iv< nvelocities; ++iv){
		    // do beamforming at this velocity and fill in beamformResult as a time series
		    beamform(traces, xcenter, ycenter, xvelocity, yvelocities[iv], beamformResult);
    	      
		    // calculate total sum of energy in this time series and record in beamformEnergy
		    double sumEnergy = 0.0;
		    for(int t=0; t<ThisFileInfo.nSamples; ++t){
		        sumEnergy += abs(beamformResult[t]);
		    }
		    beamformEnergy[iv] = sumEnergy;
		    // do some output while testing code
		    //cout << yvelocities[iv] << "\t" << beamformEnergy[iv] << "\n";
		}

		
		// record beamform energy for each velocity in the same line of the output text file
		for(int iv=0; iv<nvelocities; ++iv){
			outfile << beamformEnergy[iv] << "\t";
		}
		outfile << "\n";

		delete[] data;
	  }
      }
	
	// cleanup: close input text file containing names of other input files and output files
	fileList.close();
	outfile.close();
	
  } catch (const std::exception& error){
  	cout << "Exception "<< error.what() << "\n";
  }
  return 0;
}
