#ifndef TRACE_HPP
#define TRACE_HPP

#include <string.h> // for memcpy
#include <algorithm> // for min
#include <iostream> // for print statements
#include "trace_data.hpp"
#include "array_ops.hpp"
#include <complex.h>
#include <fftw3.h>
#include <math.h>
using namespace std;

/* This is the trace class, containing metadata about each trace and pointers to a time-domain signal container and a fourier-domain signal container, both boost multi_arrays */

class trace{
  private:
    short int sampleFrq; // number of samples per second
    int nSamples; // total number of samples in trace
    double x; // x position of receiver 
    double y; // y position of receiver
    bool timeUpToDate; // true if time series has been updated following most recent spectral alteration
    bool fourierUpToDate; // true if fourier data/spectrum has been updated following most recent time alteration
    fftw_complex * spectrum; // spectrum of time series data
    double * trace_data; // time series data
 
  public:
    // constructors (for varying amounts of data provided
    trace() { }; // empty one

    trace(short int frq, int n, double rx, double ry, double *data){
      // number of samples per second, number of samples in trace
      sampleFrq = frq;
      nSamples = n;
      // receiver location
      x = rx;
      y = ry;
      // copy data into this object's data array
      trace_data = data;
      ////trace_data = new double[n];
      ////memcpy(trace_data, data, n*sizeof(double));
      timeUpToDate = true;
      // allocate space for spectrum but dont calculate it
      spectrum = (fftw_complex *) fftw_malloc( sizeof(fftw_complex)*(1+nSamples/2));
      fourierUpToDate = false;
    };
    
    // deep copy constructor of a trace
    trace(const trace & otherTrace){
    	sampleFrq = otherTrace.sampleFrq;
    	nSamples = otherTrace.nSamples;
    	x = otherTrace.x;
    	y = otherTrace.y;
    	timeUpToDate = otherTrace.timeUpToDate;
    	fourierUpToDate = otherTrace.fourierUpToDate;
    	// allocate and do deep copy of time series data
    	trace_data = new double[nSamples];
    	memcpy(trace_data, otherTrace.trace_data, nSamples*sizeof(double));
    	// allocate and do deep copy of spectrum
    	spectrum = (fftw_complex *) fftw_malloc( sizeof(fftw_complex)*(1+nSamples/2));
    	memcpy(spectrum, otherTrace.spectrum, (1+nSamples/2)*sizeof(fftw_complex));
    };

    // print basic info about the trace (samples, frq, receiver position)
    void printTrace(){
      if(not timeUpToDate){ updateTime(); }
      std::cout << "I am a trace with " << nSamples << " samples \n";
      std::cout << "taken at a rate of "<< sampleFrq << " samples per sec \n";
      std::cout << "in position x = "<< x << " y = "<< y << "\n";
      std::cout << " first ten entries are \n";
      for(int i=0; i<std::min(10,nSamples); ++i){
	    std::cout << trace_data[i]<< "\n";
      }
    };
    
    // print basic info about the spectrum (Hz per bin, max Hz, up to 
    // 10 samples starting at designated frequency)
    void printSpectrum(float startPrintingHz){
    	if(not fourierUpToDate){ updateSpec(); }
    	float maxFrq = float(sampleFrq)/2.0;
 		float nHzPerBin = maxFrq/float(1+nSamples/2);   	
    	std::cout << "I am the spectrum of a trace with " << nHzPerBin << " Hz per bin \n";
    	std::cout << "With a maximum frequency of " << maxFrq << " Hz \n";
    	std::cout << "Starting with "<< startPrintingHz << " Hz \n";
    	std::cout << "the next 10 bins are \n";
    	int startBin = indexFromHz(startPrintingHz);
    	for(int i=startBin; i<std::min(startBin+10,1+nSamples/2); ++i){
    	  std::cout << spectrum[i][0] << " + "<< spectrum[i][1]<< " i " << " \n";
    	}
    }

	// return number of frequency Hz per bin in the spectrum
	float getHzPerBin(){
		float nBins = 1.0 + float(nSamples)/2.0; // number of bins in spectrum
      	return (float(sampleFrq)/2.0)/nBins; // Hz per interval in spectrum
	};
	
    // return the number of samples per second
    int getSampleFrq(){
      return sampleFrq;
    };

    // return the number of samples
    int getNSamples(){
      return nSamples;
    };
    
    // grab a value from the time series data at index sampleIndex
    double getSample(int sampleIndex){
    	return trace_data[sampleIndex];
    };
    
    // add a double called toAdd to the time series sample at index sampleIndex
    void addToSample(int sampleIndex, double toAdd){
    	trace_data[sampleIndex] += toAdd;
    };

    // calculate the mean of the time series trace 
    double getMean(){
      double avg=0;
      for(int i=0; i<nSamples; ++i){
	    avg += trace_data[i];
      }
      avg /= double(nSamples);
      return avg;
    };
    
    // get x position
    double getx(){
    	return x;
    };
    
    // get y position
    double gety(){
    	return y;
    };

    // get distance from (x0,y0)
    double getDistanceFrom(double x0, double y0){
      	return std::sqrt((x-x0)*(x-x0) + (y-y0)*(y-y0));
    };

    // calculate the spectrum from the time series data
    void updateSpec(){
      fftw_plan real_plan;
      real_plan = fftw_plan_dft_r2c_1d(nSamples, trace_data, spectrum, FFTW_ESTIMATE);
      fftw_execute(real_plan);
      fftw_destroy_plan(real_plan);
      // record that the Fourier transform is up to date with the most recent time series changes
      fourierUpToDate = true;
    };


    // calculate the time series data from the spectrum
    void updateTime(){
      fftw_plan inv_plan;
      inv_plan = fftw_plan_dft_c2r_1d(nSamples, spectrum, trace_data, FFTW_ESTIMATE);
      fftw_execute(inv_plan);
      fftw_destroy_plan(inv_plan);
      // scale time series by factor of nSamples so ifft(fft(time series)) is identity
      double scaleFactor = 1.0/double(nSamples);
      scaleArray(trace_data, nSamples,scaleFactor);
      // record that the time series is up to date with the most recent frequency domain changes
      timeUpToDate = true;
    };

    // return the index in the spectrum corresponding to a particular frequency in Hz
    int indexFromHz(float frequency){
      float nBins = 1.0 + float(nSamples)/2.0; // number of bins in spectrum
      float HzPerBin = (float(sampleFrq)/2.0)/nBins; // Hz per interval in spectrum
      int index = int(frequency/HzPerBin);
      if (index > 1+nSamples/2){
		std::cout << "WARNING: out of bounds frequency requested \n";
	  }      
      return int(frequency/HzPerBin);
    };
    
    // get rid of outliers in amplitude within a small window
    void despike(int nSamplesInWindow){
    	// find the mean absolute value in each window
    	int N = nSamplesInWindow/2;
    	double sum = 0.0;
		int currentWindowWidth = 0;
		double * meanAbsVal = new double[nSamples];
		for(int i=0; i<N; ++i){
			sum += std::abs(trace_data[i]);
			currentWindowWidth += 1;
		}
		for(int i=0; i<nSamples; ++i){		
			// if you aren't really close to the end of the array, add on a new element
			if(i < nSamples-N){ 
				sum += std::abs(trace_data[i+N]);
				currentWindowWidth += 1;
			}			
			// if you aren't really close to the beginning of the array drop the earliest element
			if(i >= N){
				sum -= std::abs(trace_data[i-N]);
				currentWindowWidth -= 1;
			}					
			// average absolute value in window
			meanAbsVal[i] = sum/currentWindowWidth; 
		}
		
		// if an element is an outlier in its window, zero it out
		for(int i=0; i<nSamples; ++i){
			if(trace_data[i] > 2.5*meanAbsVal[i]){
				trace_data[i] = 0.0;
			}
		}
		
		delete[] meanAbsVal; //cleanup
    };


    // whiten most of the spectrum of a signal
    void spectralWhiten(float propWhite, float minFrq, float maxFrq){
      // Within the band (minFrq Hz, maxFrq Hz) whiten the parts of the spectrum with
      // a significant amount of energy. 
      // propWhite is the proportion (in [0,1]) of the larger frequencies in the band 
      // that get whitened.
      if ((propWhite <= 1.0) && (propWhite >= 0.0)){
      	// make sure spectrum is up to date
	    if (not fourierUpToDate){ updateSpec();} 

	    // calc start and end indices of whitening
	    int startInd = indexFromHz(minFrq);  // lowest index in spectrum to whiten
	    int endInd = indexFromHz(maxFrq); // highest index in spectrum to whiten

	    // get propWhite percentile of data
	    double smallestWhite = absQuantile(&(spectrum[startInd]), endInd-startInd, propWhite); 

		// scale each piece of the spectrum
	    for(int i=startInd; i<=endInd; ++i){
	      double scaleFactor = max(smallestWhite,sqrt(spectrum[i][0]*spectrum[i][0] + spectrum[i][1]*spectrum[i][1]));
	      spectrum[i][0] = spectrum[i][0]/scaleFactor;
	      spectrum[i][1] = spectrum[i][1]/scaleFactor;
	    }
	    
	    timeUpToDate =  false; // spectrum was updated but not time series
      } else{ cout << "ERROR: need 0.0 <= propWhite <= 1.0 in trace.hpp whiten() \n";}
    };  


    // zero out spectrum in some window from minFrqHz to maxFrq Hz
    void zeroSpec(float minFrq, float maxFrq){
      // set the spectrum to zero for any frequencies in the band from minFrq Hz to maxFrq Hz
      
      if(not fourierUpToDate){ updateSpec();}
      
      int startInd = indexFromHz(minFrq); // lowest index to zero
      int endInd = indexFromHz(maxFrq); // highest index after zero
      for(int j=startInd; j<endInd; ++j){
	    spectrum[j][0] = 0.0;
	    spectrum[j][1] = 0.0;
      }
      
      timeUpToDate = false; // spectrum was updated but not time series
    }; 


	// add a constant value to each entry of the time series data 
	void addConstToTrace(double theConstant){
		if(not timeUpToDate){ updateTime(); } // update time series if it isn't up to date	
		addToArray(trace_data, nSamples, theConstant); // add theConstant to all nSamples entries of trace_data	
		fourierUpToDate = false; // spectrum wasn't updated even though time series was	
	};


	// running-absolute value mean normalization with window width equal to 
	// 0.5/minFrqOfInterest seconds, where it is suggested that minFrqOfInterest
	// should be the minimum frequency in the bandpass filter (in Hz)
	void tempNormalize(float minFrqOfInterest){
		if(not timeUpToDate){ updateTime(); } // update the time series if it isn't up to date

		// use window of width = half the max period of the bandpass filter
		float windowWidthSec = 0.5/minFrqOfInterest;
		int windowWidthIndices = int(windowWidthSec*sampleFrq); 
		int N = (windowWidthIndices-1)/2;
		
		// Start the set up of the running average then add on elements as you move along 
		// the vector. Also drop old elements as you move along the vector.
		double sum = 0.0;
		int currentWindowWidth = 0;
		double * scale_factor = new double[nSamples];
		for(int i=0; i<N; ++i){
			sum += std::abs(trace_data[i]);
			currentWindowWidth += 1;
		}
		for(int i=0; i<nSamples; ++i){		
			// if you aren't really close to the end of the array, add on a new element
			if(i < nSamples-N){ 
				sum += std::abs(trace_data[i+N]);
				currentWindowWidth += 1;
			}			
			// if you aren't really close to the beginning of the array drop the earliest element
			if(i >= N){
				sum -= std::abs(trace_data[i-N]);
				currentWindowWidth -= 1;
			}					
			// average absolute value in window
			scale_factor[i] = sum/currentWindowWidth; 
		}
		
		// actually divide each value by the avg absolute value in its window
		for(int i=0; i<windowWidthIndices; ++i){
			trace_data[i] /= scale_factor[i];
		}
			
		// cleanup 
		delete[] scale_factor;	
	};
    

    // do a cross correlation with this as the virtual source (otherReceiver uses complex conjugate)
    // and write the spectrum of the cross correlation in xcorrSpectrum
    void crossCorrelateSpec(trace & otherReceiver, fftw_complex * xcorrSpectrum){
    
      // make sure the spectra are up to date
      if(not fourierUpToDate){ updateSpec();}
      if(not otherReceiver.fourierUpToDate){ otherReceiver.updateSpec();}
      //std::cout << "before multiplying \n";
      // for each frequency multiply this spectrum by the complex conjugate of the other spectrum
      for(int i=0; i<1+nSamples/2; ++i){
	//std::cout << i << "\n";
      	xcorrSpectrum[i][0] = (spectrum[i][0] * otherReceiver.spectrum[i][0]) + (spectrum[i][1] * otherReceiver.spectrum[i][1]);
        xcorrSpectrum[i][1] = (-1.0 * spectrum[i][0] * otherReceiver.spectrum[i][1]) + (spectrum[i][1] * otherReceiver.spectrum[i][0]);
      } 
    };  


  // Deconvolution of this trace and another trace (the virtual source) and 
  // write the spectrum of the deconvolution in deconSpectrum.
  // If the other trace's spectrum in a frequency bin is below 
  // noiseParam*mean(abs(other trace's spectrum over all frqs)), that
  // frequency bin is boosted up to this threshold.noiseParam should be much 
  // less than 1, but bigger than 0.
  void deconSpec(trace & otherReceiver, fftw_complex * deconSpectrum, double noiseParam){
    // make sure the spectra are up to date
    if(not fourierUpToDate){ updateSpec();}
    if(not otherReceiver.fourierUpToDate){ otherReceiver.updateSpec();}

    // calculate the average frequency bin's amplitude
    double absMean = 0;
    double * srcAmp = new double[1+nSamples/2];
    for(int i=0; i<1+nSamples/2; ++i){
      srcAmp[i] = sqrt(otherReceiver.spectrum[i][0]*otherReceiver.spectrum[i][0] + otherReceiver.spectrum[i][1]*otherReceiver.spectrum[i][1]);
      absMean += srcAmp[i];
    }
    absMean /= (1+nSamples/2);
    double threshold = noiseParam*absMean;
    
    // do the deconvolution against the virtual source
    for(int i=0; i<1+nSamples/2; ++i){
      if(srcAmp[i] < threshold){
	double scaleFactor = threshold/(srcAmp[i]+ threshold/0.000001);
	deconSpectrum[i][0] = (spectrum[i][0]*otherReceiver.spectrum[i][0] + spectrum[i][1]*otherReceiver.spectrum[i][1])/(srcAmp[i]*srcAmp[i]*scaleFactor);
	deconSpectrum[i][1] = (spectrum[i][1]*otherReceiver.spectrum[i][0] + spectrum[i][0]*otherReceiver.spectrum[i][1])/(srcAmp[i]*srcAmp[i]*scaleFactor);
      } else{
        deconSpectrum[i][0] = (spectrum[i][0]*otherReceiver.spectrum[i][0] + spectrum[i][1]*otherReceiver.spectrum[i][1])/(srcAmp[i]*srcAmp[i]);
	deconSpectrum[i][1] = (spectrum[i][1]*otherReceiver.spectrum[i][0] - spectrum[i][0]*otherReceiver.spectrum[i][1])/(srcAmp[i]*srcAmp[i]);
      } 
    }
   
    delete[] srcAmp; 
  }


    // cross coherence of this trace and another trace and write the 
    //  spectrum of the cross coherence in xcoherSpectrum
    void crossCoherSpec(trace & otherReceiver, fftw_complex * xcoherSpectrum){
      // make sure the spectra are up to date
      if(not fourierUpToDate){ updateSpec();}
      if(not otherReceiver.fourierUpToDate){ otherReceiver.updateSpec();}
      for(int i=0; i<1+nSamples/2; ++i){
	double normalization = sqrt(spectrum[i][0]*spectrum[i][0] + spectrum[i][1]*spectrum[i][1]) * sqrt(otherReceiver.spectrum[i][0] * otherReceiver.spectrum[i][0] + otherReceiver.spectrum[i][1] * otherReceiver.spectrum[i][1]);
	normalization = max(normalization,1e-9); // 1e-9
        xcoherSpectrum[i][0] = ((spectrum[i][0] * otherReceiver.spectrum[i][0]) + (spectrum[i][1] * otherReceiver.spectrum[i][1]))/normalization;
	xcoherSpectrum[i][1] = ((-1.0 * spectrum[i][0] * otherReceiver.spectrum[i][1]) + (spectrum[i][1] * otherReceiver.spectrum[i][0]))/normalization;
      }
    }
    
    
    // return a cross correlation with this as the virtual source (otherReceiver uses complex conjugate),
    // write the spectrum of the cross correlation in xcorrSpectrum,
    // write the time series of the cross correlation in xcorrTime
    void crossCorrelate(trace & otherReceiver, fftw_complex * xcorrSpectrum, double * xcorrTime){    
	  crossCorrelateSpec(otherReceiver, xcorrSpectrum); // do the cross correlation in the frequency domain
	  // do an inverse fft for a real time series output
	  fftw_plan inv_plan;
      inv_plan = fftw_plan_dft_c2r_1d(nSamples, xcorrSpectrum, xcorrTime, FFTW_ESTIMATE);
      fftw_execute(inv_plan);
      fftw_destroy_plan(inv_plan);    
      // put in scale factor that's always in inverse fft's
      double scaleFactor = 1.0/double(nSamples);
      scaleArray(xcorrTime, nSamples,scaleFactor);  	
    };  
    
    // return the spectrum within just a limited frequency bandwidth including lowFrq and highFrq
    void getLimitedSpec(float lowFrq, int nFrq, fftw_complex * limitedSpec){
    	float maxFrq = float(sampleFrq)/2.0;
 	float nHzPerBin = maxFrq/float(1+nSamples/2);
    	int minFrqInd = int(lowFrq/nHzPerBin);
    	for(int i=0; i<nFrq; ++i){
	  limitedSpec[i][0] = spectrum[i+minFrqInd][0];
	  limitedSpec[i][1] = spectrum[i+minFrqInd][1];
    	}
    }


    // destructor
    ~trace(){
      fftw_free((void *)spectrum);
      // **** should check with deep copy constructor******
      ////if(trace_data){
      ////  delete[] trace_data;
      ////}
    };
};




 

#endif
