#include <iostream>
#include <fstream>
#include <string>
#include <assert.h>
#include <vector>
#include <cstdlib>
#include <complex>
#include <fftw3.h>
#include "reading.hpp"
#include "trace_data.hpp"
#include "trace.hpp"
#include "filter_workflows.hpp"
using namespace std;

int main(int argc, char* argv[]) {
  try{
    // input file names
    char *windowListName = argv[1]; // input file with list of file names to use
    ifstream windowList(windowListName);
    string fileinfoStr; // used as buffer for data read in from file
    string startOfFileNames = argv[2]; // start of input data file names


    // channels of interest
    int startChannel = atoi(argv[3]);
    int endChannel = atoi(argv[4]); 
    float channelSpacing = 1.0; // 1 meter per channel in this data subset (**should have header***)
    int nChannels = endChannel-startChannel+1;
    int srcChannel =  atoi(argv[8]); //int((endChannel+startChannel)/2.0);
    // number of time lags of interest
    int nLags = atoi(argv[5]);
    // frequencies (Hz) of bandpass 
    float minFrq = 3.0;
    float maxFrq = 50.0;
    
    // create array to hold cross-correlations (fast direction is nLags long)
    double * xcorrs = new double[nLags*nChannels];
    for(int i=0; i<nLags*nChannels; ++i){
      xcorrs[i] = 0;
    }

    
    // open the output text file where beamforming results will be printed
    string outfileNameStart = argv[6];
    string outfileNameEnd = argv[7];

 

    string oldInfileName;
    header_vals ThisFileInfo;


    if(windowList.is_open()){ 
      getline(windowList,fileinfoStr); // read first line and don't do anything with it
	int windowCtr = 0;
        while(getline(windowList,fileinfoStr,'\t') ){
	  // increase window number
	  windowCtr += 1;
	  // get the name of the file where data is stored
	  string infileName = startOfFileNames+fileinfoStr;
	  // get the starting second of the interval
	  getline(windowList,fileinfoStr,'\t');
	  int startSec = atoi(fileinfoStr.c_str());
	  // get the ending second of the interval
	  getline(windowList,fileinfoStr,'\n');
	  int endSec = atoi(fileinfoStr.c_str());
	  cout << infileName << "\t" << startSec << "\t" << endSec << "\n";
	    
	
	  // read relevant traces from this file
	  const char *filename = infileName.c_str();
	  ThisFileInfo = SEGYInfo(filename); // get nTraces, sampleFrq, nSamples and store in ThisFileInfo
	  int windowSamples = (ThisFileInfo.sampleFrq)*(endSec-startSec);
	  cout << ThisFileInfo.nSamples << "\t" << ThisFileInfo.sampleFrq << "\t" << windowSamples << "\t" << startSec << "\t" << endSec << "\n";
	  int startSample = (ThisFileInfo.sampleFrq)*startSec;
	  vector<trace*> traces; // creates a vector of pointers to traces
	  traces.reserve(nChannels); // preallocate space
	  double * data = new double[nChannels*windowSamples];
	  for(int channel = startChannel; channel < startChannel+nChannels; ++channel){
		    // read a single trace
		    readTrace(filename, &data[(channel-startChannel)*windowSamples], channel,ThisFileInfo.nSamples,startSample,windowSamples);
		    float y = float(channel);
		    traces.push_back(new trace(ThisFileInfo.sampleFrq, windowSamples, 0.0, y, &data[(channel-startChannel)*windowSamples]));

		    filterProc1(traces.back(),minFrq, maxFrq);
	  }
	  // ******* add in FK filter here *******
	  /*double MetersPerMile = 1609.34; 
	  double HoursPerSec = 1.0/3600.0;
	  double minVelMPerSec = 5*MetersPerMile*HoursPerSec; // min velocity of cars
	  double maxVelMPerSec = 65*MetersPerMile*HoursPerSec; // max velocity of cars
	  FK_filter(data,minVelMPerSec,maxVelMPerSec,ThisFileInfo,channelSpacing,nChannels);*/

	  // cross correlate each receiver against the receiver acting as virtual source
	  fftw_complex * xcorrSpec = (fftw_complex *)fftw_malloc( sizeof(fftw_complex)*(1+windowSamples/2));
	  double * xcorrTime = new double[windowSamples];

	       
	  for(int channel=startChannel; channel<=endChannel; ++channel){
		  traces[channel-startChannel]->crossCorrelateSpec(*traces[srcChannel-startChannel],xcorrSpec); // cross-corr frq domain
		  fftw_plan inv_plan = fftw_plan_dft_c2r_1d(windowSamples,xcorrSpec,xcorrTime,FFTW_ESTIMATE);
		  fftw_execute(inv_plan); // turn frq domain xcorr into time domain
		   // add the cross-correlations onto the xcorr stack 
		   int startPos=0;
		   for(int l=0; l<nLags/2; ++l){
		     xcorrs[l+(channel-startChannel)*nLags] += xcorrTime[l+windowSamples-nLags/2];
		   }
		   for(int l=0; l<nLags/2; ++l){
		     xcorrs[(nLags/2)+l+(channel-startChannel)*nLags] += xcorrTime[startPos+l];
		   }
		   fftw_destroy_plan(inv_plan);		   
	  }

	  // clean up traces
          delete[] xcorrTime;
	  fftw_free((void *)xcorrSpec); 
	  for(vector<trace*>::iterator iter=traces.begin(); iter != traces.end(); ++iter){
		  delete (*iter);
 	  }
	  delete[] data;

	  oldInfileName = infileName;

	  // record the cross correlation for this time window
	  char windowCtrChar [64];
	  sprintf(windowCtrChar, "%i", windowCtr);
	  string windowCtrStr = string(windowCtrChar);
	  string outfileName = outfileNameStart + windowCtrStr + outfileNameEnd;
	  ofstream outfile(outfileName.c_str(), ofstream::binary);
	  for(int channel=startChannel; channel<=endChannel; ++channel){
	    for(int lag=0; lag<nLags; ++lag){
	      outfile.write((const char*)(&xcorrs[nLags*(channel-startChannel)+lag]),sizeof(double));
	    }
	  }
	  outfile.close();
        } // end of while loop that keeps grabbing more data files

        windowList.close();

      } // end of if statement checking open files

    delete[] xcorrs;
	
  } catch (const std::exception& error){
  	cout << "Exception "<< error.what() << "\n";
  }
  return 0;
}
